@@ -1,7 +1,11 @@
 import time
 
 from hwpack.better_tarfile import writeable_tarfile
-from hwpack.packages import get_packages_file
+from hwpack.packages import (
+    FetchedPackage,
+    get_packages_file,
+    PackageMaker,
+    )
 
 
 class Metadata(object):
@@ -15,7 +19,8 @@ class Metadata(object):
 
     :ivar name: the name of the hardware pack.
     :type name: str
-    :ivar version: the version of the hardware pack.
+    :ivar version: the version of the hardware pack. It must not contain white
+        spaces.
     :type version: str
     :ivar origin: the origin of the hardware pack, or None if the origin
         is not known.
@@ -35,6 +40,10 @@ class Metadata(object):
         See the instance variables for a description of the arguments.
         """
         self.name = name
+        if ' ' in version:
+            raise AssertionError(
+                'Hardware pack version must not contain white '
+                'spaces: "%s"' % version)
         self.version = version
         self.origin = origin
         self.maintainer = maintainer
@@ -86,6 +95,7 @@ class HardwarePack(object):
     :type FORMAT: str
     """
 
+    # The format version cannot contain white spaces. 
     FORMAT = "1.0"
     FORMAT_FILENAME = "FORMAT"
     METADATA_FILENAME = "metadata"
@@ -105,7 +115,7 @@ class HardwarePack(object):
         self.sources = {}
         self.packages = []
 
-    def filename(self):
+    def filename(self, extension=".tar.gz"):
         """The filename that this hardware pack should have.
 
         Returns the filename that the hardware pack should have, according
@@ -118,9 +128,9 @@ class HardwarePack(object):
             support_suffix = ""
         else:
             support_suffix = "_%s" % self.metadata.support
-        return "hwpack_%s_%s_%s%s.tar.gz" % (
+        return "hwpack_%s_%s_%s%s%s" % (
             self.metadata.name, self.metadata.version,
-            self.metadata.architecture, support_suffix)
+            self.metadata.architecture, support_suffix, extension)
 
     def add_apt_sources(self, sources):
         """Add APT sources to the hardware pack.
@@ -153,6 +163,29 @@ class HardwarePack(object):
         """
         self.packages += packages
 
+    def add_dependency_package(self, packages_spec):
+        """Add a packge that depends on packages_spec to the hardware pack.
+
+        :param packages_spec: A list of apt package specifications,
+            e.g. ``['foo', 'bar (>= 1.2)']``.
+        """
+        with PackageMaker() as maker:
+            dep_package_name = 'hwpack-' + self.metadata.name
+            relationships = {}
+            if packages_spec:
+                relationships = {'Depends': ', '.join(packages_spec)}
+            deb_file_path = maker.make_package(
+                dep_package_name, self.metadata.version,
+                relationships, self.metadata.architecture)
+            self.packages.append(FetchedPackage.from_deb(deb_file_path))
+
+    def manifest_text(self):
+        manifest_content = ""
+        for package in self.packages:
+            manifest_content += "%s=%s\n" % (
+                package.name, package.version)
+        return manifest_content
+
     def to_file(self, fileobj):
         """Write the hwpack to a file object.
 
@@ -175,16 +208,13 @@ class HardwarePack(object):
             tf.create_file_from_string(
                 self.METADATA_FILENAME, str(self.metadata))
             tf.create_dir(self.PACKAGES_DIRNAME)
-            manifest_content = ""
             for package in self.packages:
                 if package.content is not None:
                     tf.create_file_from_string(
                         self.PACKAGES_DIRNAME + "/" + package.filename,
                         package.content.read())
-                manifest_content += "%s=%s\n" % (
-                    package.name, package.version)
             tf.create_file_from_string(
-                self.MANIFEST_FILENAME, manifest_content)
+                self.MANIFEST_FILENAME, self.manifest_text())
             tf.create_file_from_string(
                 self.PACKAGES_FILENAME,
                 get_packages_file(
