@@ -1,13 +1,23 @@
+import hashlib
+import logging
 import os
+import re
 import shutil
+from string import Template
+import subprocess
 import tempfile
 
 from apt.cache import Cache
 from apt.package import FetchError
 import apt_pkg
 
+from debian.debfile import DebFile
 
-def get_packages_file(packages, extra_text=None):
+
+logger = logging.getLogger(__name__)
+
+
+def get_packages_file(packages, extra_text=None, rel_to=None):
     """Get the Packages file contents indexing `packages`.
 
     :param packages: the packages to index.
@@ -15,6 +25,9 @@ def get_packages_file(packages, extra_text=None):
     :param extra_text: extra text to insert in to each stanza.
          Should not end with a newline.
     :type extra_text: str or None
+    :param rel_to: If present, generate the Filename: parts of the Packages
+        file as paths relative to this location.  If not present, Filename:
+        will just include the file name (not the path).
     :return: the Packages file contents indexing `packages`.
     :rtype: str
     """
@@ -25,7 +38,11 @@ def get_packages_file(packages, extra_text=None):
         if extra_text is not None:
             parts.append(extra_text)
         parts.append('Version: %s' % package.version)
-        parts.append('Filename: %s' % package.filename)
+        if rel_to is not None:
+            filename = os.path.relpath(package.filepath, rel_to)
+        else:
+            filename = package.filename
+        parts.append('Filename: %s' % filename)
         parts.append('Size: %d' % package.size)
         parts.append('Architecture: %s' % package.architecture)
         if package.depends:
@@ -73,9 +90,14 @@ def stringify_relationship(pkg, relationship):
             for or_alternative in or_dep.or_dependencies:
                 suffix = ""
                 if or_alternative.relation:
-                    suffix = " (%s %s)" % (
-                        or_alternative.relation,
-                        or_alternative.version)
+                    relation = or_alternative.relation
+                    if relation in ('<', '>'):
+                        # The choice made here by python-apt is to report the
+                        # relationship in a Python spelling; as far as apt
+                        # knows, < is a deprecated spelling of <=; << is the
+                        # spelling of "strictly less than".  Similarly for >.
+                        relation *= 2
+                    suffix = " (%s %s)" % (relation, or_alternative.version)
                 or_list.append("%s%s" % (or_alternative.name, suffix))
             relationship_list.append(" | ".join(or_list))
         relationship_str = ", ".join(relationship_list)
@@ -111,6 +133,111 @@ class DummyProgress(object):
         pass
 
 
+class TemporaryDirectoryManager(object):
+    def __init__(self):
+        self._temporary_directories = None
+
+    def __enter__(self):
+        if self._temporary_directories is not None:
+            raise AssertionError("__enter__ must not be called twice")
+        self._temporary_directories = []
+        return self
+
+    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
+        if self._temporary_directories is None:
+            return
+        for tmpdir in self._temporary_directories:
+            shutil.rmtree(tmpdir)
+        self._temporary_directories = None
+        return False
+
+    def make_temporary_directory(self):
+        """Create a temporary directory and return its path.
+
+        The created directory will be deleted on __exit__.
+        """
+        if self._temporary_directories is None:
+            raise AssertionError("__enter__ must be called")
+        tmpdir = tempfile.mkdtemp()
+        self._temporary_directories.append(tmpdir)
+        return tmpdir
+
+
+class LocalArchiveMaker(TemporaryDirectoryManager):
+
+    def sources_entry_for_debs(self, local_debs, label=None):
+        tmpdir = self.make_temporary_directory()
+        with open(os.path.join(tmpdir, 'Packages'), 'w') as packages_file:
+            packages_file.write(get_packages_file(local_debs, rel_to=tmpdir))
+        if label:
+            subprocess.check_call(
+                ['apt-ftparchive',
+                 '-oAPT::FTPArchive::Release::Label=%s' % label,
+                 'release',
+                 tmpdir],
+                stdout=open(os.path.join(tmpdir, 'Release'), 'w'))
+        return 'file://%s ./' % (tmpdir, )
+
+
+class PackageMaker(TemporaryDirectoryManager):
+    """An object that can create binary debs on the fly.
+
+    PackageMakers implement the context manager protocol to manage the
+    temporary directories the debs are created in.
+    """
+
+    # This template (and the code that uses it) is made more awkward by the
+    # fact that blank lines are invalid in control files -- so in particular
+    # when there are no relationships, there must be no blank line between the
+    # Maintainer and the Description.
+    control_file_template = Template('''\
+Package: ${name}
+Version: ${version}
+Architecture: ${architecture}
+Maintainer: Nobody
+${relationships}\
+Description: Dummy package to install a hwpack
+ This package was created automatically by linaro-media-create
+''')
+
+    def make_package(self, name, version, relationships, architecture='all'):
+        tmp_dir = self.make_temporary_directory()
+        filename = '%s_%s_%s' % (name, version, architecture)
+        packaging_dir = os.path.join(tmp_dir, filename)
+        os.mkdir(packaging_dir)
+        os.mkdir(os.path.join(packaging_dir, 'DEBIAN'))
+        relationship_strs = []
+        for relationship_name, relationship_value in relationships.items():
+            relationship_strs.append(
+                '%s: %s\n' % (relationship_name, relationship_value))
+        subst_vars = dict(
+            architecture=architecture,
+            name=name,
+            relationships=''.join(relationship_strs),
+            version=version,
+            )
+        control_file_text = self.control_file_template.safe_substitute(
+            subst_vars)
+        with open(os.path.join(
+            packaging_dir, 'DEBIAN', 'control'), 'w') as control_file:
+            control_file.write(control_file_text)
+        proc = subprocess.Popen(
+            ['dpkg-deb', '-b', packaging_dir],
+            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        (stdoutdata, stderrdata) = proc.communicate()
+        if proc.returncode:
+            raise ValueError("dpkg-deb failed!\n%s" % stderrdata)
+        if stderrdata:
+            raise ValueError("dpkg-deb had warnings:\n%s" % stderrdata)
+        deb_file_path_match = re.match(
+            "dpkg-deb: building package `.*' in `(.*)'", stdoutdata)
+        if not deb_file_path_match:
+            raise ValueError(
+                "failed to find filename in dpkg-deb output:\n%s"
+                % stdoutdata)
+        return deb_file_path_match.group(1)
+
+
 class FetchedPackage(object):
     """The result of fetching packages.
 
@@ -183,6 +310,14 @@ class FetchedPackage(object):
         self.replaces = replaces
         self.breaks = breaks
         self.content = None
+        self._file_path = None
+
+    @property
+    def filepath(self):
+        if self._file_path is not None:
+            return self._file_path
+        else:
+            return self.filename
 
     @classmethod
     def from_apt(cls, pkg, filename, content=None):
@@ -217,30 +352,64 @@ class FetchedPackage(object):
             pkg.content = content
         return pkg
 
+    @classmethod
+    def from_deb(cls, deb_file_path):
+        """Create a FetchedPackage from a binary package on disk."""
+        debcontrol = DebFile(deb_file_path).control.debcontrol()
+        name = debcontrol['Package']
+        version = debcontrol['Version']
+        filename = os.path.basename(deb_file_path)
+        size = os.path.getsize(deb_file_path)
+        md5sum = hashlib.md5(open(deb_file_path).read()).hexdigest()
+        architecture = debcontrol['Architecture']
+        depends = debcontrol.get('Depends')
+        pre_depends = debcontrol.get('Pre-Depends')
+        conflicts = debcontrol.get('Conflicts')
+        recommends = debcontrol.get('Recommends')
+        provides = debcontrol.get('Provides')
+        replaces = debcontrol.get('Replaces')
+        breaks = debcontrol.get('Breaks')
+        pkg = cls(
+            name, version, filename, size, md5sum, architecture, depends,
+            pre_depends, conflicts, recommends, provides, replaces, breaks)
+        pkg.content = open(deb_file_path)
+        pkg._file_path = deb_file_path
+        return pkg
+
+    # A list of attributes that are compared to determine equality.  Note that
+    # we don't include the contents here -- we assume that comparing the md5
+    # checksum is enough (more philosophically, FetchedPackages are equal if
+    # they represent the same underlying package, even if they represent it in
+    # slightly different ways)
+    _equality_attributes = (
+        'name',
+        'version',
+        'filename',
+        'size',
+        'md5',
+        'architecture',
+        'depends',
+        'pre_depends',
+        'conflicts',
+        'recommends',
+        'provides',
+        'replaces',
+        'breaks')
+
+    @property
+    def _equality_data(self):
+        return tuple(
+            getattr(self, attr) for attr in self._equality_attributes)
+
     def __eq__(self, other):
-        def get_content(pkg):
-            return pkg.content and pkg.content.read()
-        content = get_content(self)
-        other_content = get_content(other)
-        return (self.name == other.name
-                and self.version == other.version
-                and self.filename == other.filename
-                and content == other_content
-                and self.size == other.size
-                and self.md5 == other.md5
-                and self.architecture == other.architecture
-                and self.depends == other.depends
-                and self.pre_depends == other.pre_depends
-                and self.conflicts == other.conflicts
-                and self.recommends == other.recommends
-                and self.provides == other.provides
-                and self.replaces == other.replaces
-                and self.breaks == other.breaks
-                )
+        return self._equality_data == other._equality_data
 
     def __ne__(self, other):
         return not self.__eq__(other)
 
+    def __hash__(self):
+        return hash(self._equality_data)
+
     def __repr__(self):
         has_content = self.content and "yes" or "no"
         return (
@@ -260,7 +429,7 @@ class IsolatedAptCache(object):
     :type cache: apt.cache.Cache
     """
 
-    def __init__(self, sources, architecture=None):
+    def __init__(self, sources, architecture=None, prefer_label=None):
         """Create an IsolatedAptCache.
 
         :param sources: a list of sources such that they can be prefixed
@@ -272,6 +441,7 @@ class IsolatedAptCache(object):
         self.sources = sources
         self.architecture = architecture
         self.tempdir = None
+        self.prefer_label = prefer_label
 
     def prepare(self):
         """Prepare the IsolatedAptCache for use.
@@ -280,6 +450,7 @@ class IsolatedAptCache(object):
         of sources.
         """
         self.cleanup()
+        logger.debug("Writing apt configs")
         self.tempdir = tempfile.mkdtemp(prefix="hwpack-apt-cache-")
         dirs = ["var/lib/dpkg",
                 "etc/apt/",
@@ -300,7 +471,16 @@ class IsolatedAptCache(object):
                 f.write(
                     'Apt {\nArchitecture "%s";\n'
                     'Install-Recommends "true";\n}\n' % self.architecture)
+        if self.prefer_label is not None:
+            apt_preferences = os.path.join(
+                self.tempdir, "etc", "apt", "preferences")
+            with open(apt_preferences, 'w') as f:
+                f.write(
+                    'Package: *\n'
+                    'Pin: release l=%s\n'
+                    'Pin-Priority: 1001\n' % self.prefer_label)
         self.cache = Cache(rootdir=self.tempdir, memonly=True)
+        logger.debug("Updating apt cache")
         self.cache.update()
         self.cache.open()
         return self
@@ -350,7 +530,7 @@ class DependencyNotSatisfied(Exception):
 class PackageFetcher(object):
     """A class to fetch packages from a defined list of sources."""
 
-    def __init__(self, sources, architecture=None):
+    def __init__(self, sources, architecture=None, prefer_label=None):
         """Create a PackageFetcher.
 
         Once created a PackageFetcher should have its `prepare` method
@@ -362,7 +542,8 @@ class PackageFetcher(object):
         :param architecture: the architecture to fetch packages for.
         :type architecture: str
         """
-        self.cache = IsolatedAptCache(sources, architecture=architecture)
+        self.cache = IsolatedAptCache(
+            sources, architecture=architecture, prefer_label=prefer_label)
 
     def prepare(self):
         """Prepare the PackageFetcher for use.
@@ -395,6 +576,7 @@ class PackageFetcher(object):
         :param packages: the list of package names to ignore.
         :type packages: an iterable of str
         """
+        logger.debug("Ignoring %s" % packages)
         for package in packages:
             self.cache.cache[package].mark_install(auto_fix=False)
             if self.cache.cache.broken_count:
@@ -413,6 +595,7 @@ class PackageFetcher(object):
             candidate = package.installed
             base = os.path.basename(candidate.filename)
             installed.append(FetchedPackage.from_apt(candidate, base))
+            logger.debug("Ignored %s" % package.name)
         self.cache.set_installed_packages(installed)
         broken = [p.name for p in self.cache.cache
                 if p.is_inst_broken or p.is_now_broken]
@@ -430,6 +613,7 @@ class PackageFetcher(object):
                 seen_packages.add(package.name)
         all_packages = set(package_dict.keys())
         for unseen_package in all_packages.difference(seen_packages):
+            logger.debug("%s is ignored, skipping" % unseen_package)
             del package_dict[unseen_package]
 
     def fetch_packages(self, packages, download_content=True):
@@ -476,6 +660,7 @@ class PackageFetcher(object):
         acq = apt_pkg.Acquire(DummyProgress())
         acqfiles = []
         for package in self.cache.cache.get_changes():
+            logger.debug("Fetching %s ..." % package)
             candidate = package.candidate
             base = os.path.basename(candidate.filename)
             if package.name not in fetched:
@@ -487,6 +672,7 @@ class PackageFetcher(object):
                 acq, candidate.uri, candidate.md5, candidate.size,
                 base, destfile=destfile)
             acqfiles.append((acqfile, result_package, destfile))
+            logger.debug(" ... from %s" % acqfile.desc_uri)
         self.cache.cache.clear()
         acq.run()
         for acqfile, result_package, destfile in acqfiles:
