@@ -1,4 +1,6 @@
-from testtools.matchers import Matcher, Mismatch
+import tarfile
+
+from testtools.matchers import Annotate, Equals, Matcher, Mismatch
 
 
 class TarfileMissingPathMismatch(Mismatch):
@@ -65,7 +67,8 @@ class TarfileHasFile(Matcher):
 
     def __init__(self, path, type=None, size=None, mtime=None,
                  mtime_skew=None, mode=None, linkname=None, uid=None,
-                 gid=None, uname=None, gname=None, content=None):
+                 gid=None, uname=None, gname=None, content=None,
+                 content_matcher=None):
         """Create a TarfileHasFile Matcher.
 
         :param path: the path that must be present.
@@ -92,6 +95,9 @@ class TarfileHasFile(Matcher):
             None to not check.
         :param content: the content that `path` must have when extracted,
             or None to not check.
+        :param content_matcher: a matcher to match the content that `path` has
+            when extracted, or None to not check.  You can't specify both
+            content_matcher and content.
         """
         self.path = path
         self.type = type
@@ -104,7 +110,18 @@ class TarfileHasFile(Matcher):
         self.gid = gid
         self.uname = uname
         self.gname = gname
-        self.content = content
+        if content is not None:
+            if content_matcher is not None:
+                raise ValueError(
+                    "doesn't make sense to specify content and "
+                    "content_matcher")
+            content_matcher = Equals(content)
+        if content_matcher is not None:
+            self.content_matcher = Annotate(
+                'The content of path "%s" did not match' % path,
+                content_matcher)
+        else:
+            self.content_matcher = None
 
     def match(self, tarball):
         """Match a tarfile.TarFile against the expected values."""
@@ -122,14 +139,26 @@ class TarfileHasFile(Matcher):
                         attr, tarball, self.path, expected, actual)
         if self.mtime is not None:
             mtime_skew = self.mtime_skew or 0
-            if abs(self.mtime - info.mtime) > self.mtime_skew:
+            if abs(self.mtime - info.mtime) > mtime_skew:
                 return TarfileWrongValueMismatch(
                     "mtime", tarball, self.path, self.mtime, info.mtime)
-        if self.content is not None:
-            actual = tarball.extractfile(self.path).read()
-            if actual != self.content:
-                return TarfileWrongValueMismatch(
-                    "content", tarball, self.path, self.content, actual)
+        if self.content_matcher is not None:
+            if info.type == tarfile.DIRTYPE:
+                contents = []
+                path_frags = self.path.split('/')
+                for name in tarball.getnames():
+                    name_frags = name.split('/')
+                    if (len(name_frags) == len(path_frags) + 1 and
+                        name_frags[:-1] == path_frags):
+                        contents.append(name_frags[-1])
+                content_mismatch = self.content_matcher.match(contents)
+                if content_mismatch:
+                    return content_mismatch
+            else:
+                actual = tarball.extractfile(self.path).read()
+                content_mismatch = self.content_matcher.match(actual)
+                if content_mismatch:
+                    return content_mismatch
         return None
 
     def __str__(self):
